"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit floors utility functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import Autodesk
from Autodesk.Revit.DB import BuiltInCategory, FilteredElementCollector, FloorType


# doc:   current model document
def _get_all_floor_types_by_category(doc):
    """
    Function returning a filtered element collector of all floor types in the model.

    This uses builtinCategory as filter. Return types includes:
    - Floor
    - In place families or loaded families

    It will therefore not return any foundation slab types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of floor types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_Floors)
        .WhereElementIsElementType()
    )
    return collector


# doc   current model document
def _get_floor_types_by_class(doc):
    """
    Function returning a filtered element collector of all floor types in the model.

    - Floor
    - Foundation Slab

    it will therefore not return any in place family types ...

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of floor types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = FilteredElementCollector(doc).OfClass(FloorType)
    return collector



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A function used to output messages to a console.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities import date_stamps as dateStamp
from duHast.Utilities import padding as pad


def default_out_print(message):
    """
    Default pip out method: print to console.

    :param message: A message
    :type message: str
    """

    print(message)


def output(message="", pipe_out=default_out_print):
    """
    Print message to pipe.

    Note:

    - If message is not a string it will convert it to a string.
    - Multi line strings will pe printed line by line.
    - strings containing [true] ot [false] will be padded to have [true] or [false] to the right hand side.

    :param message: The message, defaults to ''
    :type message: str, optional
    """

    # make sure message is a string:
    if type(message) != str:
        message = str(message)

    # check for multi row messages
    if "\n" in message:
        message_chunks = message.split("\n")
        for message_chunk in message_chunks:
            pipe_out("{}".format(pad.pad_string(message_chunk)))
    else:
        pipe_out("{}".format(pad.pad_string(message)))


def output_with_time_stamp(message="", pipe_out=default_out_print):
    """
    Print message to console.

    Note:

    - The message will be prefixed with a date stamp in format '2022-08-09 19_09_19 :'
    - If message is not a string it will convert it to a string.
    - Multi line strings will pe printed line by line
    - strings containing [true] ot [false] will be padded to have [true] or [false] to the right hand side.

    :param message: The message, defaults to ''
    :type message: str, optional
    """

    # make sure message is a string:
    if type(message) != str:
        message = str(message)

    # get the current time string
    # 2022-08-09 19:09:19 :
    timestamp = dateStamp.get_date_stamp("%y-%m-%d %H_%M_%S :")

    # check for multi row messages
    if "\n" in message:
        message_chunks = message.split("\n")
        for message_chunk in message_chunks:
            pipe_out("{} {}".format(timestamp, pad.pad_string(message_chunk)))
    else:
        pipe_out("{} {}".format(timestamp, pad.pad_string(message)))


def output_header(message="", pipe_out=default_out_print):
    """
    Prints header message to pipe.

    Note:

    - If message is not a string it will convert it to a string.

    :param message: The message, defaults to ''
    :type message: str, optional
    """

    # make sure message is a string:
    if type(message) != str:
        message = str(message)

    pipe_out("{}".format(pad.pad_header_no_time_stamp(message)))


def output_header_with_time_stamp(message="", pipe_out=default_out_print):
    """
    Prints header message to pipe.

    Note:

    - If message is not a string it will convert it to a string.
    - The header will be prefixed with a date stamp in format '2022-08-09 19_09_19 :'

    :param message: The message, defaults to ''
    :type message: str, optional
    """

    # make sure message is a string:
    if type(message) != str:
        message = str(message)

    pipe_out("{}".format(pad.pad_header_with_time_stamp(message)))



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A module with helper function around family types catalogue files.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.family_types_get_data_from_xml import get_type_data_via_XML_from_family_file

def export_catalogue_file(doc, file_path = None):
    """
    Export the family types catalogue file.

    :param doc: The family document to extract the type data from.
    :type doc: rdb.Family
    :param file_path: The path to save the catalogue file. If None, the file will be saved in the same location as the family with the same name as the family.
    :type file_path: str
    """

    # check if a family document...


    # get the family name
    family_name = doc.Name

    # remove the file extension
    if family_name.lower().endswith(".rfa"):
        family_name = family_name[:-4]

    # get the family path
    family_path = doc.PathName

    # get the root path (same as the family path)
    root_path = doc.Name

    # get the root category path (same as the family category)
    root_category_path = doc.FamilyCategory.Name

    # get the family type data
    family_type_data = get_type_data_via_XML_from_family_file(doc, family_name, family_path, root_path, root_category_path)

    # convert type data into catalogue file

    


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data base class for Revit object properties.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.base import Base


class DataBase(Base):
    def __init__(self, data_type, **kwargs):
        """
        Class constructor

        :param data_type: human readable data type
        :type data_type: str
        """
        
        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(DataBase, self).__init__(**kwargs)
        self.data_type = data_type
        if isinstance(self.data_type, str) == False:
            raise TypeError(
                "data_type must be a string, got {} instead.".format(type(data_type))
            )

    @property
    def DataType(self):
        """
        Property: returns the data type of this class.

        :return:  A string representing the data type
        :rtype: str
        """

        return self.data_type

    def __eq__(self, other):
        if not isinstance(other, DataBase):
            return NotImplemented
        return self.DataType == other.DataType

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def __hash__(self):
        return hash(self.DataType)



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit stair run elements. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb
import Autodesk.Revit.DB.Architecture as rdbA

from duHast.Revit.Common import common as com

#: list of built in parameters for stair run type
STAIR_RUN_TYPE_PARAS = [rdb.BuiltInParameter.STAIRSTYPE_RUN_TYPE]


def get_stair_run_types_by_class(doc):
    """
    Gets a filtered element collector of all Stair run types in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector containing stair run types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return rdb.FilteredElementCollector(doc).OfClass(rdbA.StairsRunType)


def get_stair_run_types_ids_by_class(doc):
    """
    Gets all Stair run element type ids available in model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing stair run types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_class = get_stair_run_types_by_class(doc)
    ids = com.get_ids_from_element_collector(col_class)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains utility function(s) for material reports. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Utilities import files_io as fileIO, utility as util
from duHast.Revit.Materials.materials import get_all_materials


def get_material_report_data(doc, revit_file_path):
    """
    Gets material data ready for being written to file.
    - HOSTFILE
    - ID
    - NAME
    - and any parameter names and values attached to a material
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The fully qualified file path of Revit file.
    :type revit_file_path: str
    :return: The material data in a nested list of string
    :rtype: list of list of str
    """

    data = []
    mats = get_all_materials(doc)
    for mat in mats:
        try:
            paras = mat.GetOrderedParameters()
            for p in paras:
                para_name = p.Definition.Name
                p_value = rParaGet.get_parameter_value(p)
                data.append(
                    [
                        revit_file_path,
                        str(mat.Id),
                        util.encode_ascii(rdb.Element.Name.GetValue(mat)),
                        util.encode_ascii(para_name),
                        util.encode_ascii(p_value),
                    ]
                )
        except Exception:
            data.append(
                [
                    fileIO.get_file_name_without_ext(revit_file_path),
                    str(mat.Id),
                    util.encode_ascii(rdb.Element.Name.GetValue(mat)),
                ]
            )
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A module with helper function around family types data extraction using Revit xml export functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Supports 2 methods of data extraction:

- from family file on disk
- from family element instance in document

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr

clr.AddReference("System.Xml")
from System.Xml import XmlDocument, XmlNamespaceManager

import tempfile
import os

from Autodesk.Revit.DB import Element

from duHast.Revit.Family.Data.Objects.family_type_parameter_data_storage import (
    FamilyTypeParameterDataStorage,
)
from duHast.Revit.Family.Data.Objects.family_type_data_storage import (
    FamilyTypeDataStorage,
)
from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import get_file_name_without_ext, get_directory_path_from_file_path


def write_data_to_temp_xml_file_and_read_it_back(an_action_to_write_xml_data):
    """
    Write the data to a temp XML file and read it back.

    :param an_action_to_write_xml_data: The action to write the XML data.
    :type an_action_to_write_xml_data: function

    :return: The data read back from the XML file.
    :rtype: XmlDocument or None if an error occurred.
    """

    doc_xml = None

    # Create a temporary file
    with tempfile.NamedTemporaryFile(delete=False, suffix=".xml") as temp_file:
        temp_path_xml = temp_file.name

    try:

        # Write the data to the file
        an_action_to_write_xml_data(temp_path_xml)

        # Read the data back from the file
        with open(temp_path_xml, "r") as file:
            xml_content = file.read()

        # Load the XML content
        doc_xml = XmlDocument()
        doc_xml.LoadXml(xml_content)

    finally:
        # Delete the temporary file
        if os.path.exists(temp_path_xml):
            os.remove(temp_path_xml)

    return doc_xml

def write_data_to_xml_file_and_read_it_back(an_action_to_write_xml_data, xml_file_path):
    """
    Write the data to an XML file and read it back.

    :param an_action_to_write_xml_data: The action to write the XML data.
    :type an_action_to_write_xml_data: function
    :param xml_file_path: The path of the XML file.
    :type xml_file_path: str

    :return: The data read back from the XML file.
    :rtype: XmlDocument or None if an error occurred.
    """

    doc_xml = None

    try:

        # Write the data to the file
        an_action_to_write_xml_data(xml_file_path)

        # Read the data back from the file
        with open(xml_file_path, "r") as file:
            xml_content = file.read()

        # Load the XML content
        doc_xml = XmlDocument()
        doc_xml.LoadXml(xml_content)
    except Exception as e:
        return None
    return doc_xml


def read_xml_into_storage(doc_xml, family_name, family_path):
    """
    Read the XML data into the storage object.

    :param doc_xml: The XML document.
    :type doc_xml: XmlDocument
    :param family_name: The name of the family.
    :type family_name: str
    :param family_path: The path of the family file.
    :type family_path: str

    :return: A list of family type data objects.
    :rtype: list[FamilyTypeDataStorage]
    """

    type_data = []
    # Add an XML namespace manager
    name_space_manager = XmlNamespaceManager(doc_xml.NameTable)
    name_space_manager.AddNamespace("atom", "http://www.w3.org/2005/Atom")
    name_space_manager.AddNamespace("A", "urn:schemas-autodesk-com:partatom")

    # get some family information i.e. the root category path
    root_category_path = "None"

    # Select the family node
    family_node = doc_xml.SelectSingleNode("//A:family", name_space_manager)

    # Get the family parameters
    for part_node in family_node.SelectNodes("A:part", name_space_manager):
        # Get the family type name
        family_type_name = None
        for child_node in part_node.ChildNodes:
            if child_node.Name == "title":
                family_type_name = child_node.InnerText
                break

        # If we got a type name, add the parameters, their values and units, parameter type and type of parameter
        if family_type_name:
            parameters = []
            for child_node in part_node.ChildNodes:
                if child_node.Name != "title":
                    
                    # attempt to read out values
                    name = "unknown type"
                    try:
                        name = child_node.Name
                    except Exception as e:
                        name ="{}: {}".format(name, e)
                    
                    type = "unknown type"
                    try:
                        type = child_node.Attributes["type"].Value
                    except Exception as e:
                        type ="{}: {}".format(type, e)
                    
                    type_of_parameter = "unknown type"
                    try:
                        type_of_parameter = child_node.Attributes["typeOfParameter"].Value
                    except Exception as e:
                        type_of_parameter ="{}: {}".format(type_of_parameter, e)
                    
                    units = "unknown type"
                    try:
                        units = child_node.Attributes["units"].Value
                    except Exception as e:
                        units ="{}: {}".format(units, e)

                    # Create a parameter object
                    parameter = FamilyTypeParameterDataStorage(
                        name=name,
                        type=type,
                        type_of_parameter=type_of_parameter,
                        units=units,
                        value=child_node.InnerText,
                    )
                    
                    # Add type to family
                    parameters.append(parameter)

            # Set up a family type data storage object
            fam_type = FamilyTypeDataStorage(
                root_name_path=family_name,
                root_category_path=root_category_path,
                family_name=family_name,
                family_file_path=family_path,
                family_type_name=family_type_name,
                parameters=parameters,
            )

            # Add the family type to the list of types
            type_data.append(fam_type)
    return type_data


def get_type_data_via_XML_from_family_file(application, family_name, family_path, use_temporary_file=True):
    """
    Get the family type data from the family document using the XML extraction method.
    This can be used to extract the type data from a family document within a Revit session but without opening the family in Revit.

    :param application: The Revit application object.
    :type application: Autodesk.Revit.ApplicationServices.Application
    :param family_name: The name of the family.
    :type family_name: str
    :param family_path: The path of the family file.
    :type family_path: str
    :param use_temporary_file: Whether to use a temporary file for the XML data.
    :type use_temporary_file: bool

    :return: A result object with .result containing a list of family type data objects. (or empty if failed)
    :rtype: Result
    """

    return_value = Result()

    # Set up list of type information to be returned
    type_data = []

    try:
        # set up action to write xml data
        def action(temp_path_xml):
            # Save XML file to temporary location
            # this is a method of the application object and does not require the family to be open...
            application.ExtractPartAtomFromFamilyFile(family_path, temp_path_xml)

        doc_xml = None

        if use_temporary_file:
            # Write the data to an XML file and read it back
            doc_xml = write_data_to_temp_xml_file_and_read_it_back(action)
        else:
            dir_out = get_directory_path_from_file_path(family_path)
            family_name = get_file_name_without_ext(family_path)

            # Write the data to an XML file and read it back
            doc_xml = write_data_to_xml_file_and_read_it_back(action, os.path.join(dir_out,family_name + ".xml"))

        # check if an xml document was created
        if doc_xml is None:
            return_value.update_sep(False, "No XML document was created.")
            return return_value

        # read the xml data into the storage object
        type_data = read_xml_into_storage(doc_xml, family_name, family_path)

        # store list in return object
        return_value.result.append(type_data)
    except Exception as e:
        return_value.update_sep(False, "{}".format(e))

    return return_value


def get_type_data_via_XML_from_family_object(revit_family):
    """
    Get the family type data from the family element in a REvit document using the XML extraction method.
   
    :param revit_family: The Revit family object.
    :type revit_family: Autodesk.Revit.DB.Family

    :return: A result object with .result containing a list of family type data objects. (or empty if failed)
    :rtype: Result
    """

    return_value = Result()
    # Set up list of type information to be returned
    type_data = []

    try:
        # set up action to write xml data
        def action(temp_path_xml):
            # Save XML file to temporary location
            revit_family.ExtractPartAtom(temp_path_xml)

        # Write the data to an XML file and read it back
        doc_xml = write_data_to_temp_xml_file_and_read_it_back(action)

        # check if an xml document was created
        if doc_xml is None:
            return_value.update_sep(False, "No XML document was created.")
            return return_value

        # read the xml data into the storage object
        type_data = read_xml_into_storage(
            doc_xml, family_name=Element.Name.GetValue(revit_family), family_path=""
        )

        # store list in return object
        return_value.result.append(type_data)
    except Exception as e:
        return_value.update_sep(False, "{}".format(e))
    
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family type data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.Family.Data.Objects.family_type_data_storage import (
    FamilyTypeDataStorage,
)
from duHast.Revit.Family.Data.Objects.family_type_data_processor_defaults import (
    NESTING_SEPARATOR,
)

from duHast.Revit.Family.Data.Objects.family_type_data_storage import (
    FamilyTypeDataStorage,
)

# import Autodesk
# import Autodesk.Revit.DB as rdb

# data dictionary key values specific to this class
CATEGORY_NAME = "categoryName"


class FamilyTypeData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None):
        """
        Class constructor

        :param rootPath: The path of the nested family in a tree: rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param rootCategoryPath: The path of the family category in a tree: rootCategoryName :: nestedCategoryNameOne :: nestedCategoryTwo\
            This includes the actual category name as the last node.
        :type rootCategoryPath: str
        """

        # store data type  in base class
        super(FamilyTypeData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
        )
        # super(CategoryData, self).__init__(rootPath, dataType)

        if root_category_path != None:
            category_chunks = root_category_path.split(NESTING_SEPARATOR)
            self.category = category_chunks[-1]
        else:
            self.category = "unknown"

        self.saved_file_name = ""

    def process(self, doc, session_id):
        """
        Collects all base data from the document and stores it in the class property .data

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        """

        # make sure to get a value for the file path which is not empty if the document has not been saved
        if doc.PathName != "":
            self.saved_file_name = doc.PathName

        # TODO: xml out may be tricky to implement, there only exist 2 functions:
        #   1. gxml from family file ( gets data from a family file on disc)
        #   2. xml from family class instance
        # Need to implement a version whicj just works with the open family document
            

        # save out xml and read family type data back in
        # types_data = get_type_data_via_XML_from_family_file(
        #     doc=doc,
        #     family_name=doc.Title,
        #     family_path=self.saved_file_name,
        #     root_path=self.root_path,
        #     root_category_path=self.root_category_path,
        # )

        # # add type data to data
        # for type_data in types_data:
        #     self.add_data(type_data)

    def get_data(self):
        return self.data

    def add_data(self, storage_instance):
        if isinstance(storage_instance, FamilyTypeDataStorage):
            self.data.append(storage_instance)
        else:
            raise ValueError(
                "storage instance must be an instance of FamilyTypeDataStorage"
            )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to comma separated text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

#import clr
import codecs
import csv
import System.IO

from duHast.Utilities.files_io import get_file_name_without_ext, remove_null_bytes, file_delete
from duHast.Utilities.Objects.result import Result


def get_unique_headers(files):
    """
    Gets a list of alphabetically sorted headers retrieved from text files.
    Assumes:

    - first row in each file is the header row
    - headers are separated by <tab> character

    :param files: List of file path from which the headers are to be returned.
    :type files: list of str
    :return: List of headers.
    :rtype: list of str
    """

    headers_in_all_files = {}
    for f in files:
        # get unmodified row data and remove the next line character at the end
        data = get_first_row_in_csv_file(f)
        headers_in_all_files[get_file_name_without_ext(f)] = data
    headers_unique = []
    for header_by_file in headers_in_all_files:
        empty_header_counter = 0
        for header in headers_in_all_files[header_by_file]:
            # reformat any empty headers to be unique
            if header == "":
                header = header_by_file + ".Empty." + str(empty_header_counter)
                empty_header_counter = empty_header_counter + 1
            if header not in headers_unique:
                headers_unique.append(header)
    return sorted(headers_unique)


def process_csv(file_path):
    """
    Process a CSV file by removing null bytes and then reading its content.

    :param file_path: The path to the CSV file to be processed.
    :type file_path: str
    """
    # Create a temporary file in the system's temp directory
    temp_file_path = System.IO.Path.GetTempFileName()
    
    row_list = []

    try:
        # Remove null bytes and save to a temporary file
        remove_null_bytes(file_path, temp_file_path)

        # Read the cleaned file with the CSV reader
        with open(temp_file_path, 'r') as cleaned_file:
            reader = csv.reader(cleaned_file)
            for row in reader:
                row_list.append(row)
            cleaned_file.close()
    finally:
        # Clean up the temporary file
        if System.IO.File.Exists(temp_file_path):
            System.IO.File.Delete(temp_file_path)
    
    return row_list


def read_csv_file_with_encoding(file_path_csv, increase_max_field_size_limit=False):
    """
    Read a csv file, attempting to detect and handle encoding, including BOMs.

    :param filepathCSV: The fully qualified file path to the csv file.
    :type filepathCSV: str
    :return: A list of list of strings representing the data in each row.
    :rtype: list of list of str
    """

    row_list = []
    encodings = ['utf-8-sig', 'utf-16']

    return_value = Result()
    if increase_max_field_size_limit:
        csv.field_size_limit(2147483647)

    for encoding in encodings:
        try:
            with codecs.open(file_path_csv, 'r', encoding=encoding) as csv_file:
                reader = csv.reader(csv_file)
                row_list = [row for row in reader]
            
            # Successful read
            return_value.append_message("read file successfully")
            return_value.status=True
            return_value.result=row_list
            return return_value
        except (Exception) as e:
            return_value.update_sep(False, "Failed with encoding {}: {}".format(encoding, e))

    # statsu should be false 
    return_value.update_sep(False, "Failed to decode using known encodings.")
    return return_value


def read_csv_file(filepathCSV, increaseMaxFieldSizeLimit=False):
    """
    Read a csv file into a list of rows, where each row is another list.

    :param filepathCSV: The fully qualified file path to the csv file.
    :type filepathCSV: str
    :return: A list of list of strings representing the data in each row.
    :rtype: list of list of str
    """

    row_list = []
    # read with encoding enabled
    read_result = read_csv_file_with_encoding(filepathCSV, increaseMaxFieldSizeLimit)

    if read_result.status:
        return read_result.result
    
    # if that failed try the below...

    # hard coded hack
    if increaseMaxFieldSizeLimit:
        csv.field_size_limit(2147483647)

    try:
        with open(filepathCSV) as csv_file:
            reader = csv.reader(csv_file)
            row_list = [row for row in reader]
            csv_file.close()
    except csv.Error as e:
        # maybe a nullbyte exception?
        if "line contains NULL byte" in str(e):
            row_list = process_csv(filepathCSV)
        else:
            row_list = []
    except Exception as e:
        row_list = []
    return row_list


def get_first_row_in_csv_file(filePath):
    """
    Reads the first line of a csv text file and returns it as a list of strings
    :param filePath: The fully qualified file path.
    :type filePath: str
    :return: The first row of a text file.
    :rtype: str
    """

    return_value = []
    try:
        with open(filePath) as csv_file:
            reader = csv.reader(csv_file)
            for row in reader:  # each row is a list
                return_value = row
                break
            csv_file.close()
    except Exception as e:
        print(str(e))
    return return_value


def write_report_data_as_csv(
    file_name, header, data, write_type="w", enforce_ascii=False,  encoding="utf-8", bom=None, quoting=csv.QUOTE_NONE, delimiter=','
):
    """
    Function writing out report information as CSV file.
    :param file_name: The reports fully qualified file path.
    :type file_name: str
    :param header: list of column headers
    :type header: list of str
    :param data: list of list of strings representing row data
    :type data: [[str,str,..]]
    :param write_type: Flag indicating whether existing report file is to be overwritten 'w' or appended to 'a', defaults to 'w'
    :type write_type: str, optional
    :param enforce_ascci: Flag to enforce ASCII encoding on data. If True, data will be encoded to ASCII. Defaults to False.
    :type enforce_ascci: bool, optional
    :param encoding: Encoding used to write the file. Defaults to 'utf-8'.
    :type encoding: str, optional
    :param bom: the byte order mark, Default is None (none will be written). BOM: "utf-16" = , "utf-16-le" = ,  utf-8 =
    :type bom: str, default is NoneType
    :param quoting: Quoting style used by the csv writer. Defaults to csv.QUOTE_NONE. Options are csv.QUOTE_ALL, csv.QUOTE_MINIMAL, csv.QUOTE_NONNUMERIC, csv.QUOTE_NONE
    :type quoting: int, optional
    """

    return_value = Result()
    
    # Open the file with the codecs.open method to specify encoding
    with codecs.open(file_name, write_type, encoding=encoding) as f:
        try:
            # Write a newline character if appending to the file to make sure we are starting on a new line
            if write_type == 'a':
                f.write('\n')
            
            # Write BOM manually if specified
            if bom and 'w' in write_type:
                f.write(bom.decode(encoding))

            # Create the CSV writer
            # line terminator is set to '\n' to avoid double newlines on Windows
            writer = csv.writer(f, delimiter=delimiter, escapechar='\\', quoting=quoting, lineterminator='\n')

            def encoded_row(row):
                if enforce_ascii:
                    return [s.encode('ascii', 'ignore').decode('ascii') for s in row]
                else:
                    return row  # Keep the strings in their current state for writing

            # Write header
            if header:
                # check if header only or if there is data to write as well
                if data:
                    # Write the header row with the CSV writer including a new line character
                    writer.writerow(encoded_row(header))
                    return_value.append_message("Header written to file. (including newline)")
                else:
                    # Write the header row without a newline character
                    f.write(delimiter.join(encoded_row(header)))
                    return_value.append_message("Header written to file. (including newline)")
            else:
                return_value.append_message("No header provided to write to file.")

            # Write data rows, looping over rows to prevent new line character on the last row
            for i in range(len(data)):
                row = data[i]
                if i == len(data) - 1:
                    # Write the last row without a newline character
                    f.write(delimiter.join(encoded_row(row)))
                    return_value.append_message("Last row {} written to file. (without newline)>>{}".format(','.join(encoded_row(row)), i))
                else:
                    writer.writerow(encoded_row(row))
                    return_value.append_message("Row {} written to file. (including newline)>>{}".format(','.join(encoded_row(row)), i))
        
        
        except Exception as e:
                return_value.update_sep(False, "File: {} failed to write data with exception: {}".format(file_name, e))    
        finally:
            # make sure to close the file
            f.close()

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of base element filter functions relating to Revit stairs. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import Autodesk
import Autodesk.Revit.DB as rdb
import Autodesk.Revit.DB.Architecture as rdbA


def _get_all_stair_types_by_category(doc):
    """
    Gets a filtered element collector of all Stair types in the model.

    Return includes:
    - Stair
    - Assembled Stair
    - Precast Stair
    - Cast-In-Place Stair
    - In place families or loaded families

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing stair types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_Stairs)
        .WhereElementIsElementType()
    )
    return collector


def _get_stair_types_by_class(doc):
    """
    Gets a filtered element collector of all Stair types in the model.

    Return includes:

    - Assembled Stair
    - Precast Stair
    - Cast-In-Place Stair

    It will not return any in place family or Stair types! These are internally treated as Families or Family Symbols class objects.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing stair types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return rdb.FilteredElementCollector(doc).OfClass(rdbA.StairsType)
